/*
 *  Copyright 2001-2005 Adrian Thurston <thurston@cs.queensu.ca>
 *            2004 Eric Ocean <eric.ocean@ampede.com>
 */

/*  This file is part of Ragel.
 *
 *  Ragel is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 * 
 *  Ragel is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with Ragel; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

#include "ragel.h"
#include "ftabcodegen.h"
#include "redfsm.h"
#include "parsetree.h"

/* Determine if we should use indicies or not. */
void FTabCodeGen::calcIndexSize()
{
	int sizeWithInds = 0, sizeWithoutInds = 0;

	/* Calculate cost of using with indicies. */
	for ( RedStateList::Iter st = redFsm->stateList; st.lte(); st++ ) {
		int totalIndex = st->outSingle.length() + st->outRange.length() + 
				(st->defTrans == 0 ? 0 : 1);
		sizeWithInds += arrayTypeSize(maxIndex) * totalIndex;
	}
	sizeWithInds += arrayTypeSize(maxState) * redFsm->transSet.length();
	if ( anyActions() )
		sizeWithInds += arrayTypeSize(maxActListId) * redFsm->transSet.length();

	/* Calculate the cost of not using indicies. */
	for ( RedStateList::Iter st = redFsm->stateList; st.lte(); st++ ) {
		int totalIndex = st->outSingle.length() + st->outRange.length() + 
				(st->defTrans == 0 ? 0 : 1);
		sizeWithoutInds += arrayTypeSize(maxState) * totalIndex;
		if ( anyActions() )
			sizeWithoutInds += arrayTypeSize(maxActListId) * totalIndex;
	}

	/* If using indicies reduces the size, use them. */
	useIndicies = sizeWithInds < sizeWithoutInds;
}

std::ostream &FTabCodeGen::TO_STATE_ACTION( RedStateAp *state )
{
	int act = 0;
	if ( state->toStateAction != 0 )
		act = state->toStateAction->actListId+1;
	out << act;
	return out;
}

std::ostream &FTabCodeGen::FROM_STATE_ACTION( RedStateAp *state )
{
	int act = 0;
	if ( state->fromStateAction != 0 )
		act = state->fromStateAction->actListId+1;
	out << act;
	return out;
}

std::ostream &FTabCodeGen::EOF_ACTION( RedStateAp *state )
{
	int act = 0;
	if ( state->eofAction != 0 )
		act = state->eofAction->actListId+1;
	out << act;
	return out;
}


/* Write out the function for a transition. */
std::ostream &FTabCodeGen::TRANS_ACTION( RedTransAp *trans )
{
	int action = 0;
	if ( trans->action != 0 )
		action = trans->action->actListId+1;
	out << action;
	return out;
}

/* Write out the function switch. This switch is keyed on the values
 * of the func index. */
std::ostream &FTabCodeGen::TO_STATE_ACTION_SWITCH()
{
	/* Loop the actions. */
	for ( ActionTableMap::Iter redAct = redFsm->actionMap; redAct.lte(); redAct++ ) {
		if ( redAct->numToStateRefs > 0 ) {
			/* Write the entry label. */
			out << "\tcase " << redAct->actListId+1 << ":\n";

			/* Write each action in the list of action items. */
			for ( ActionTable::Iter item = redAct->key; item.lte(); item++ )
				ACTION( item->value, 0, false );

			out << "\tbreak;\n";
		}
	}

	/* Write the directive for going back into the output file. The line
	 * number is for the next line, so add one. */
	out << "#line " << outFilter->line + 1 << " \""; LDIR_PATH(outputFile) << "\"\n";
	return out;
}

/* Write out the function switch. This switch is keyed on the values
 * of the func index. */
std::ostream &FTabCodeGen::FROM_STATE_ACTION_SWITCH()
{
	/* Loop the actions. */
	for ( ActionTableMap::Iter redAct = redFsm->actionMap; redAct.lte(); redAct++ ) {
		if ( redAct->numFromStateRefs > 0 ) {
			/* Write the entry label. */
			out << "\tcase " << redAct->actListId+1 << ":\n";

			/* Write each action in the list of action items. */
			for ( ActionTable::Iter item = redAct->key; item.lte(); item++ )
				ACTION( item->value, 0, false );

			out << "\tbreak;\n";
		}
	}

	/* Write the directive for going back into the output file. The line
	 * number is for the next line, so add one. */
	out << "#line " << outFilter->line + 1 << " \""; LDIR_PATH(outputFile) << "\"\n";
	return out;
}

std::ostream &FTabCodeGen::EOF_ACTION_SWITCH()
{
	/* Loop the actions. */
	for ( ActionTableMap::Iter redAct = redFsm->actionMap; redAct.lte(); redAct++ ) {
		if ( redAct->numEofRefs > 0 ) {
			/* Write the entry label. */
			out << "\tcase " << redAct->actListId+1 << ":\n";

			/* Write each action in the list of action items. */
			for ( ActionTable::Iter item = redAct->key; item.lte(); item++ )
				ACTION( item->value, 0, true );

			out << "\tbreak;\n";
		}
	}
	out << "#line " << outFilter->line + 1 << " \""; LDIR_PATH(outputFile) << "\"\n";
	return out;
}

/* Write out the function switch. This switch is keyed on the values
 * of the func index. */
std::ostream &FTabCodeGen::ACTION_SWITCH()
{
	/* Loop the actions. */
	for ( ActionTableMap::Iter redAct = redFsm->actionMap; redAct.lte(); redAct++ ) {
		if ( redAct->numTransRefs > 0 || redAct->numEofRefs > 0 ) {
			/* Write the entry label. */
			out << "\tcase " << redAct->actListId+1 << ":\n";

			/* Write each action in the list of action items. */
			for ( ActionTable::Iter item = redAct->key; item.lte(); item++ )
				ACTION( item->value, 0, false );

			out << "\tbreak;\n";
		}
	}

	/* Write the directive for going back into the output file. The line
	 * number is for the next line, so add one. */
	out << "#line " << outFilter->line + 1 << " \""; LDIR_PATH(outputFile) << "\"\n";
	return out;
}

std::ostream &FTabCodeGen::STATE_DATA()
{
	out << 
		"static "; ARRAY_TYPE(maxKeyOffset) << " "; FSM_NAME() << "_ko[] = {\n";
		KEY_OFFSETS() <<
		"};\n"
		"\n";

	out <<
		"static "; ALPH_TYPE() << " "; FSM_NAME() << "_k[] = {\n";
		KEYS() <<
		"};\n"
		"\n";

	out << 
		"static "; ARRAY_TYPE(maxSingleLen) << " "; FSM_NAME() << "_sl[] = {\n";
		SINGLE_LENS() <<
		"};\n"
		"\n";

	out << 
		"static "; ARRAY_TYPE(maxRangeLen) << " "; FSM_NAME() << "_rl[] = {\n";
		RANGE_LENS() <<
		"};\n"
		"\n";

	out << 
		"static "; ARRAY_TYPE(maxIndexOffset) << " "; FSM_NAME() << "_io[] = {\n";
		INDEX_OFFSETS() <<
		"};\n"
		"\n";

	if ( useIndicies ) {
		out << 
			"static "; ARRAY_TYPE(maxIndex) << " "; FSM_NAME() << "_i[] = {\n";
			INDICIES() <<
			"};\n"
			"\n";

		out <<
			"static "; ARRAY_TYPE(maxState) << " "; FSM_NAME() << "_tt[] = {\n";
			TRANS_TARGS_WI() <<
			"};\n"
			"\n";

		if ( anyActions() ) {
			out <<
				"static "; ARRAY_TYPE(maxActListId) << " "; FSM_NAME() << "_ta[] = {\n";
				TRANS_ACTIONS_WI() <<
				"};\n"
				"\n";
		}
	}
	else {
		out <<
			"static "; ARRAY_TYPE(maxState) << " "; FSM_NAME() << "_tt[] = {\n";
			TRANS_TARGS() <<
			"};\n"
			"\n";

		if ( anyActions() ) {
			out <<
				"static "; ARRAY_TYPE(maxActListId) << " "; FSM_NAME() << "_ta[] = {\n";
				TRANS_ACTIONS() <<
				"};\n"
				"\n";
		}
	}

	if ( anyToStateActions() ) {
		out <<
			"static "; ARRAY_TYPE(maxActionLoc) << " "; FSM_NAME() << "_tsa[] = {\n";
			TO_STATE_ACTIONS() <<
			"};\n"
			"\n";
	}

	if ( anyFromStateActions() ) {
		out <<
			"static "; ARRAY_TYPE(maxActionLoc) << " "; FSM_NAME() << "_fsa[] = {\n";
			FROM_STATE_ACTIONS() <<
			"};\n"
			"\n";
	}

	if ( anyEofActions() ) {
		out <<
			"static "; ARRAY_TYPE(maxActListId) << " "; FSM_NAME() << "_ea[] = {\n";
			EOF_ACTIONS() <<
			"};\n"
			"\n";
	}

	out <<
		"static int "; FSM_NAME() << "_start = "; START_STATE_ID() << ";\n"
		"\n";

	return out;
}


/* Init base class. */
CFTabCodeGen::CFTabCodeGen( char *fsmName, ParseData *parseData, 
		RedFsmAp *redFsm, std::ostream &out )
:
	FsmCodeGen(fsmName, parseData, redFsm, out)
{ }

void CFTabCodeGen::writeOutCode()
{
	/* Write out arrays of states/transitions. */
	STATE_DATA();

	/* Init routine. */
	INIT_ROUTINE();

	/* The binary search. */
	BSEARCH() << "\n";

	/* The binary search for a range. */
	RANGE_BSEARCH() << "\n";

	/* Execution function. */
	out << 
		"int "; FSM_NAME() << "_execute( struct "; FSM_NAME() << " *fsm, "; EL_TYPE() <<
				" *_data, int _len )\n"
		"{\n"
		"	"; EL_TYPE() << " *_p = _data-1;\n"
		"	"; EL_TYPE() << " *_pe = _data+_len;\n"
		"	int _slen, _rlen, _cs = fsm->curs";

	if ( anyRegCurStateRef() )
		out << ", _ps";

	if ( anyActions() ) 
		out << ", _acts";

	out <<
		";\n"
		"	"; ALPH_TYPE() << " *_keys;\n"
		"	int _trans;\n"
		"	fsm->curs = -1;\n"
		"\n";

	EXECUTE_BEGIN();

	out << 
		"	goto resume;\n"
		"\n";

	/* Head of the loop. */
	out << "again:\n";

	if ( anyToStateActions() ) {
		out <<
			"	_acts = "; FSM_NAME() << "_tsa[_cs];\n"
			"	switch ( _acts ) {\n";
			TO_STATE_ACTION_SWITCH() <<
			"	}\n"
			"\n";
	}

	out <<
		"resume:\n"
		"	if ( ++_p == _pe )\n"
		"		goto out;\n"
		"\n";

	if ( anyFromStateActions() ) {
		out <<
			"	_acts = "; FSM_NAME() << "_fsa[_cs];\n"
			"	switch ( _acts ) {\n";
			FROM_STATE_ACTION_SWITCH() <<
			"	}\n"
			"\n";
	}

	LOCATE_TRANS() <<
		"\n"
		"match:\n";

	if ( anyRegCurStateRef() )
		out << "	_ps = _cs;\n";

	if ( useIndicies ) {
		out << "	_trans = "; 
		FSM_NAME() << "_i[_trans];\n";
	}

	out <<
		"	_cs = "; FSM_NAME() << "_tt[_trans];\n"
		"\n";

	if ( anyActions() ) {
		out << 
			"	if ( "; FSM_NAME() << "_ta[_trans] == 0 )\n"
			"		goto again;\n"
			"	_acts = "; FSM_NAME() << "_ta[_trans];\n"
			"\n";

		out <<
			"	switch ( _acts ) {\n";
			ACTION_SWITCH() <<
			"	}\n"
			"\n";
	}

	out <<
		"	goto again;\n"
		"\n";

	out << 
		"out:\n"
		"	fsm->curs = _cs;\n";

	if ( redFsm->errState != 0 ) {
		out <<
			"	if ( _cs == "; ERROR_STATE() << " )\n"
			"		return -1;\n";
	}

	out << 
		"	if ( _cs >= "; FIRST_FINAL_STATE() << " )\n"
		"		return 1;\n"
		"	return 0;\n"
		"}\n"
		"\n";

	/* Finish routine. */
	out <<
		"int "; FSM_NAME() << "_finish( struct "; FSM_NAME() << " *fsm )\n"
		"{\n";

	if ( anyEofActions() ) {
		if ( anyEofActionCharRef() ) {
			out << "	"; EL_TYPE();
			if ( ! anyEofActionHold() )
				out << " *_p = 0;\n";
			else {
				out << " *_p = (("; EL_TYPE() << "*)0)-1" << 
						", *_pe = (("; EL_TYPE() << "*)0)+1;\n";
			}
		}

		out << 
			"	int _cs = fsm->curs;\n"
			"	fsm->curs = -1;\n";

		if ( anyEofActionHold() ) {
			out <<
				"again:\n"
				"	if ( ++_p == _pe )\n"
				"		goto out;\n";
		}

		out <<
			"	switch ( "; FSM_NAME() << "_ea[_cs] ) {\n";
			EOF_ACTION_SWITCH() <<
			"	}\n"
			"\n";

		if ( anyEofActionHold() )
			out << "	goto again;\n";

		if ( anyEofActionControl() || anyEofActionHold() )
			out << "out:\n";

		out << "	fsm->curs = _cs;\n";
	}

	if ( redFsm->errState != 0 ) {
		out <<
			"	if ( fsm->curs == "; ERROR_STATE() << " )\n"
			"		return -1;\n";
	}

	out <<
		"	if ( fsm->curs >= "; FIRST_FINAL_STATE() << " )\n"
		"		return 1;\n"
		"	return 0;\n"
		"}\n"
		"\n";

	for ( ContextMap::Iter ctx = parseData->contextMap; ctx.lte(); ctx++ ) {
		out << "unsigned char _"; FSM_NAME() << "_ctxdata_" << ctx->key << "[] = {\n\t";
		CONTEXT( ctx->value->id );
		out << "};\n\n";
	}
}


/* Init base data. */
CppFTabCodeGen::CppFTabCodeGen( char *fsmName, ParseData *parseData, 
		RedFsmAp *redFsm, std::ostream &out )
:
	FsmCodeGen(fsmName, parseData, redFsm, out) 
{ }

void CppFTabCodeGen::writeOutCode()
{
	/* Write out arrays of states/transitions. */
	STATE_DATA();

	/* Init routine. */
	INIT_ROUTINE();

	/* The binary search. */
	BSEARCH() << "\n";

	/* The binary search for a range. */
	RANGE_BSEARCH() << "\n";

	/* Execution function. */
	out << 
		"int "; FSM_NAME() << "::execute( "; EL_TYPE() << " *_data, int _len )\n"
		"{\n"
		"	"; EL_TYPE() << " *_p = _data-1;\n"
		"	"; EL_TYPE() << " *_pe = _data+_len;\n"
		"	int _slen, _rlen, _cs = this->curs";

	if ( anyRegCurStateRef() )
		out << ", _ps";
	
	if ( anyActions() ) 
		out << ", _acts";

	out <<
		";\n"
		"	"; ALPH_TYPE() << " *_keys;\n"
		"	int _trans;\n"
		"	this->curs = -1;\n"
		"\n";

	EXECUTE_BEGIN();

	out << 
		"	goto resume;\n"
		"\n";

	/* Head of the loop. */
	out << "again:\n";

	if ( anyToStateActions() ) {
		out <<
			"	_acts = "; FSM_NAME() << "_tsa[_cs];\n"
			"	switch ( _acts ) {\n";
			TO_STATE_ACTION_SWITCH() <<
			"	}\n"
			"\n";
	}

	out <<
		"resume:\n"
		"	if ( ++_p == _pe )\n"
		"		goto out;\n"
		"\n";

	if ( anyFromStateActions() ) {
		out <<
			"	_acts = "; FSM_NAME() << "_fsa[_cs];\n"
			"	switch ( _acts ) {\n";
			FROM_STATE_ACTION_SWITCH() <<
			"	}\n"
			"\n";
	}

	LOCATE_TRANS() <<
		"\n"
		"match:\n";
	
	if ( anyRegCurStateRef() )
		out << "	_ps = _cs;\n";

	if ( useIndicies ) {
		out << "	_trans = "; 
		FSM_NAME() << "_i[_trans];\n";
	}

	out <<
		"	_cs = "; FSM_NAME() << "_tt[_trans];\n"
		"\n";

	if ( anyActions() ) {
		out << 
			"	if ( "; FSM_NAME() << "_ta[_trans] == 0 )\n"
			"		goto again;\n"
			"	_acts = "; FSM_NAME() << "_ta[_trans];\n"
			"\n";

		out << 
			"	switch ( _acts ) {\n";
			ACTION_SWITCH() <<
			"	}\n"
			"\n";
	}

	out <<
		"	goto again;\n"
		"\n";

	out <<
		"out:\n"
		"	this->curs = _cs;\n";

	if ( redFsm->errState != 0 ) {
		out <<
			"	if ( _cs == "; ERROR_STATE() << " )\n"
			"		return -1;\n";
	}
	
	out <<
		"	if ( _cs >= "; FIRST_FINAL_STATE() << " )\n"
		"		return 1;\n"
		"	return 0;\n"
		"}\n"
		"\n";

	/* Finish routine. */
	out <<
		"int "; FSM_NAME() << "::finish( )\n"
		"{\n";

	if ( anyEofActions() ) {
		if ( anyEofActionCharRef() ) {
			out << "	"; EL_TYPE();
			if ( ! anyEofActionHold() )
				out << " *_p = 0;\n";
			else {
				out << " *_p = (("; EL_TYPE() << "*)0)-1" << 
						", *_pe = (("; EL_TYPE() << "*)0)+1;\n";
			}
		}

		out << 
			"	int _cs = this->curs;\n"
			"	this->curs = -1;\n";

		if ( anyEofActionHold() ) {
			out <<
				"again:\n"
				"	if ( ++_p == _pe )\n"
				"		goto out;\n";
		}

		out <<
			"	switch ( "; FSM_NAME() << "_ea[_cs] ) {\n";
			EOF_ACTION_SWITCH() <<
			"	}\n"
			"\n";

		if ( anyEofActionHold() )
			out << "	goto again;\n";

		if ( anyEofActionControl() || anyEofActionHold() )
			out << "out:\n";

		out << "	this->curs = _cs;\n";
	}

	if ( redFsm->errState != 0 ) {
		out <<
			"	if ( this->curs == "; ERROR_STATE() << " )\n"
			"		return -1;\n";
	}

	out <<
		"	if ( this->curs >= "; FIRST_FINAL_STATE() << " )\n"
		"		return 1;\n"
		"	return 0;\n"
		"}\n"
		"\n";

	for ( ContextMap::Iter ctx = parseData->contextMap; ctx.lte(); ctx++ ) {
		out << "unsigned char "; FSM_NAME() << "::_ctxdata_" << ctx->key << "[] = {\n\t";
		CONTEXT( ctx->value->id );
		out << "};\n\n";
	}
}


/* Init base data. */
ObjCFTabCodeGen::ObjCFTabCodeGen( char *fsmName, ParseData *parseData, 
		RedFsmAp *redFsm, std::ostream &out )
:
	FsmCodeGen(fsmName, parseData, redFsm, out) 
{ }

void ObjCFTabCodeGen::writeOutCode()
{
	/* Write out arrays of states/transitions. */
	STATE_DATA();

	INIT_ROUTINE();

	/* The binary search. */
	BSEARCH() << "\n";

	/* The binary search for a range. */
	RANGE_BSEARCH() << "\n";

	/* Execution function. */
	out <<
		"- (int) executeWithData:("; EL_TYPE() << " *)_data len:(int)_len;\n"
		"{\n"
		"\t"; EL_TYPE() << " *_p = _data - 1;\n"
		"\t"; EL_TYPE() << " *_pe = _data + _len;\n"
		"\tint _slen, _rlen, _cs = self->curs";

	if ( anyRegCurStateRef() )
		out << ", _ps";
	
	if ( anyActions() ) 
		out << ", _acts";

	out <<
		";\n"
		"\t"; ALPH_TYPE() << " *_keys;\n"
		"\tint _trans;\n"
		"\tself->curs = -1;\n"
		"\n";

	EXECUTE_BEGIN();

	out << 
		"	goto resume;\n"
		"\n";

	/* Head of the loop. */
	out << "again:\n";

	if ( anyToStateActions() ) {
		out <<
			"	_acts = "; FSM_NAME() << "_tsa[_cs];\n"
			"	switch ( _acts ) {\n";
			TO_STATE_ACTION_SWITCH() <<
			"	}\n"
			"\n";
	}

	out <<
		"resume:\n"
		"	if ( ++_p == _pe )\n"
		"		goto out;\n"
		"\n";

	if ( anyFromStateActions() ) {
		out <<
			"	_acts = "; FSM_NAME() << "_fsa[_cs];\n"
			"	switch ( _acts ) {\n";
			FROM_STATE_ACTION_SWITCH() <<
			"	}\n"
			"\n";
	}

	LOCATE_TRANS() <<
		"\n"
		"match:\n";
	
	if ( anyRegCurStateRef() )
		out << "\t_ps = _cs;\n";

	if ( useIndicies ) {
		out << "\t_trans = "; 
		FSM_NAME() << "_i[_trans];\n";
	}

	out <<
			"\t_cs = "; FSM_NAME() << "_tt[_trans];\n"
		"\n";

	if ( anyActions() ) {
		out << 
				"\tif ( "; FSM_NAME() << "_ta[_trans] == 0 )\n"
					"\t\tgoto again;\n"
				"\t_acts = "; FSM_NAME() << "_ta[_trans];\n"
			"\n";

		out << 
				"\tswitch ( _acts ) {\n";
				ACTION_SWITCH() <<
				"\t}\n"
			"\n";
	}

	out <<
			"\tgoto again;\n"
		"\n";

	out <<
		"out:\n"
			"\tself->curs = _cs;\n";

	if ( redFsm->errState != 0 ) {
		out <<
			"\tif ( _cs == "; ERROR_STATE() << " ) return -1;\n";
	}
	
	out <<
			"\treturn ( _cs >= "; FIRST_FINAL_STATE() << " ) ? 1 : 0;\n"
		"}\n"
		"\n";

	out <<
		"- (int) finish;\n"
		"{\n";
	
	if ( anyEofActions() ) {
		if ( anyEofActionCharRef() ) {
			out << "	"; EL_TYPE();
			if ( ! anyEofActionHold() )
				out << " *_p = 0;\n";
			else {
				out << " *_p = (("; EL_TYPE() << "*)0)-1" << 
						", *_pe = (("; EL_TYPE() << "*)0)+1;\n";
			}
		}

		out << 
			"	int _cs = self->curs;\n"
			"	self->curs = -1;\n";

		if ( anyEofActionHold() ) {
			out <<
				"again:\n"
				"	if ( ++_p == _pe )\n"
				"		goto out;\n";
		}

		out <<
			"	switch ( "; FSM_NAME() << "_ea[_cs] ) {\n";
			EOF_ACTION_SWITCH() <<
			"	}\n"
			"\n";

		if ( anyEofActionHold() )
			out << "	goto again;\n";

		if ( anyEofActionControl() || anyEofActionHold() )
			out << "out:\n";

		out << "	self->curs = _cs;\n";
	}

	if ( redFsm->errState != 0 ) {
		out <<
			"	if ( self->curs == "; ERROR_STATE() << " ) return -1;\n";
	}

	out <<
		"	return ( self->curs >= "; FIRST_FINAL_STATE() << " ) ? 1 : 0;\n";

	out <<
		"}\n"
		"\n";

	for ( ContextMap::Iter ctx = parseData->contextMap; ctx.lte(); ctx++ ) {
		out << "unsigned char "; FSM_NAME() << "::_ctxdata_" << ctx->key << "[] = {\n\t";
		CONTEXT( ctx->value->id );
		out << "};\n\n";
	}
}
